use super::{AccountName, UnitName};

use chumsky::prelude::*;

#[derive(Clone, Copy, Hash, PartialEq, PartialOrd, Debug, Ord, Eq)]
pub enum Direction {
    Deposit,
    Withdrawal,
}

#[derive(Clone, Debug, PartialEq, PartialOrd, Ord, Eq)]
pub struct Balance {
    pub account: AccountName,
    pub amount: usize,
    pub dir: Direction,
    pub unit: UnitName,
}

#[derive(Clone, Debug, PartialEq, PartialOrd, Eq, Ord)]
pub struct LedgerEntry {
    pub datestamp: (u16, u8, u8),
    pub title: Option<String>,
    pub balances: Vec<Balance>,
}

impl LedgerEntry {
    pub fn modifies(&self, account: AccountName) -> bool {
        self.balances
            .iter()
            .find(|b| b.account == account)
            .is_some()
    }
}

fn ledger_parser<'a>()
-> impl Parser<'a, &'a str, Vec<LedgerEntry>, chumsky::extra::Err<chumsky::error::Rich<'a, char>>> {
    let int = chumsky::text::digits(10)
        .collect()
        .map(|v: String| v.parse::<usize>().unwrap());

    let datestamp = group((int, just('-').ignored(), int, just('-').ignored(), int))
        .map(|(y, _, m, _, d)| (y as u16, m as u8, d as u8));

    let mark = |m| {
        chumsky::text::inline_whitespace()
            .ignore_then(just(m))
            .then_ignore(chumsky::text::inline_whitespace())
    };

    let balance = group((
        mark('-'),
        chumsky::text::ident().map(|v| stringstore::StoredString::new(v)),
        mark(':'),
        choice((
            mark('-').map(|_| Direction::Withdrawal),
            mark('+').map(|_| Direction::Deposit),
        )),
        int,
        just('.'),
        int,
        chumsky::text::inline_whitespace(),
        chumsky::primitive::none_of("\n")
            .repeated()
            .collect::<String>(),
        chumsky::text::newline(),
    ))
    .map(|(_, acc, _, dir, w, _, f, _, unit, _)| Balance {
        account: acc,
        dir,
        amount: w,
        unit: "UNIT".into(),
    });

    let entry = group((
        chumsky::text::whitespace(),
        datestamp,
        mark(':'),
        chumsky::text::inline_whitespace(),
        chumsky::primitive::none_of("\n")
            .repeated()
            .collect::<String>(),
        chumsky::text::newline(),
        balance.repeated().at_least(1).collect(),
        chumsky::text::whitespace(),
    ))
    .map(|(_, datestamp, _, _, title, _, balances, _)| LedgerEntry {
        datestamp,
        title: (!title.is_empty()).then_some(title),
        balances,
    });

    entry.repeated().collect()
}

pub fn parse_ledger(
    source: super::SourceFile,
    data: &str,
) -> Result<Vec<LedgerEntry>, super::DataError> {
    let parser = ledger_parser();

    let (presult, errors) = parser.parse(data).into_output_errors();

    if let Some(e) = errors.first() {
        let span = e.span().start()..e.span().end();

        let report = ariadne::Report::build(ariadne::ReportKind::Error, (source, span.clone()))
            .with_label(ariadne::Label::new((source, span)).with_message(e.reason()))
            .finish();

        Err(crate::data::DataError::Report(report))
    } else {
        Ok(presult.unwrap())
    }
}
